<?php

/*
 * PHP include file for admin pages for Routes plugin.
 */

/**
 * Puts a message on the screen that update info is available elsewhere.
 */
function routes_plugin_row($file, $plugin_data) {
  global $routesFileName;

  if ($file == $routesFileName) {
    echo '<tr><td colspan="5" class="plugin-update">';
    _e("No information on Routes updates. Check the plugin home page, or subscribe to its RSS feed.", 'routes');
    echo "</td></tr>\n";
  }
}

/**
 * Activation function -- creates or updates database tables for plugin.
 */
function routes_install_tables() {
  global $routesDBVersion;
  global $routesWaypointTable;
  global $routesPOITable;
  global $routesIconTable;
  // Not currently being used, but leave it installed anyway for now...
  global $routesPostMapTypeTable;
  global $routesPostIconTable;
  global $routesPostLineStyleTable;
  global $wpdb;

  // newer versions of WP have this file in a different place...

  if (file_exists(ABSPATH . 'wp-admin/includes/upgrade.php')) {
    require_once(ABSPATH . 'wp-admin/includes/upgrade.php');
  }
  else {
    require_once(ABSPATH . 'wp-admin/upgrade-functions.php');
  }

  // Note: dbDelta is picky! Don't try to alter this formatting.
  $tableStructure = "CREATE TABLE " . $routesWaypointTable . " (" .
    "waypoint_id BIGINT(20) NOT NULL AUTO_INCREMENT, \n" .
    "post_id BIGINT(20) NOT NULL, \n" .
    "waypoint_order DOUBLE, \n" .
    "address VARCHAR(100), \n" .
    "city VARCHAR(50), \n" .
    "state_country VARCHAR(50), \n" .
    "latitude VARCHAR(10), \n" .
    "longitude VARCHAR(10), \n" .
    "elevation VARCHAR(10), \n" .
    "tag TEXT, \n" .
    "PRIMARY KEY  (waypoint_id), \n" .
    "KEY post_id (post_id), \n" .
    "KEY waypoint_order (waypoint_order), \n" .
    "KEY city (city), \n" .
    "KEY state_country (state_country));";

  dbDelta($tableStructure, true);

  // Note: dbDelta is picky! Don't try to alter this formatting.
  $tableStructure = "CREATE TABLE " . $routesIconTable . " (" .
    "icon_id BIGINT(20) NOT NULL AUTO_INCREMENT, \n" .
    "admin_name VARCHAR(200) NOT NULL, \n" .
    "legend_name VARCHAR(200), \n" .
    "icon_image VARCHAR(200) NOT NULL, \n" .
    "shadow_image VARCHAR(200), \n" .
    "size_x MEDIUMINT, \n" .
    "size_y MEDIUMINT, \n" .
    "shadow_size_x MEDIUMINT, \n" .
    "shadow_size_y MEDIUMINT, \n" .
    "anchor_x MEDIUMINT, \n" .
    "anchor_y MEDIUMINT, \n" .
    "shadow_anchor_x MEDIUMINT, \n" .
    "shadow_anchor_y MEDIUMINT, \n" .
    "PRIMARY KEY  (icon_id), \n" .
    "KEY legend_name (legend_name), \n" .
    "KEY admin_name (admin_name));";

  dbDelta($tableStructure, true);

  // Note: dbDelta is picky! Don't try to alter this formatting.
  // icon type:
  // 1 = for aggregate map
  // 2 = for start of route
  // 3 = for tagged point
  $tableStructure = "CREATE TABLE " . $routesPostIconTable . " (" .
    "post_icon_id BIGINT(20) NOT NULL AUTO_INCREMENT, \n" .
    "post_id BIGINT(20) NOT NULL, \n" .
    "icon_id BIGINT(20) NOT NULL, \n" .
    "icon_type TINYINT, \n" .
    "PRIMARY KEY  (post_icon_id), \n" .
    "KEY post_id (post_id), \n" .
    "KEY icon_type (icon_type), \n" .
    "KEY icon_id (icon_id));";

  dbDelta($tableStructure, true);

  // Note: dbDelta is picky! Don't try to alter this formatting.
  $tableStructure = "CREATE TABLE " . $routesPostLineStyleTable . " (" .
    "post_id BIGINT(20) NOT NULL, \n" .
    "color CHAR(6) NOT NULL, \n" .
    "opacity VARCHAR(10), \n" .
    "weight MEDIUMINT, \n" .
    "PRIMARY KEY  (post_id));";

  dbDelta($tableStructure, true);

  // Note: dbDelta is picky! Don't try to alter this formatting.
  $tableStructure = "CREATE TABLE " . $routesPostMapTypeTable . " (" .
    "post_id BIGINT(20) NOT NULL, \n" .
    "maptype MEDIUMINT, \n" .
    "PRIMARY KEY  (post_id));";

  dbDelta($tableStructure, true);

  // Note: dbDelta is picky! Don't try to alter this formatting.
  $tableStructure = "CREATE TABLE " . $routesPOITable . " (" .
    "poi_id BIGINT(20) NOT NULL AUTO_INCREMENT, \n" .
    "post_id BIGINT(20) NOT NULL, \n" .
    "address VARCHAR(100), \n" .
    "latitude VARCHAR(10), \n" .
    "longitude VARCHAR(10), \n" .
    "icon_id BIGINT(20) NOT NULL, \n" .
    "tag TEXT, \n" .
    "PRIMARY KEY  (poi_id), \n" .
    "KEY post_id (post_id));";

  dbDelta($tableStructure, true);

  // Add the default line style, if the table is empty.
  $qry = "SELECT * FROM $routesPostLineStyleTable";
  $styles = $wpdb->get_results($qry);
  if (!isset($styles) || !count($styles)) {
    routes_save_post_line_style(-1, array('0000FF', '0.5', 3));
  }

  add_option("routes_db_version", $routesDBVersion);
  update_option("routes_db_version", $routesDBVersion);
}

/**
 * Sets up admin screen for plugin, and post editing screen section.
 *
 * @internal
 */
function routes_admin_screen() {
  global $routesFileName;

  add_options_page(__('Routes Options', 'routes'),
                    __('Routes', 'routes'), 6,
                    'routes_options',
                    'routes_options_page');

  add_options_page(__('Routes Icons', 'routes'),
                    __('Routes Icons', 'routes'), 6,
                    'routes_icons',
                    'routes_icons_page');

  add_meta_box('routeswaypoints', __('Routes Waypoints', 'routes'),
    'routes_waypoint_inner_edit_form', 'post', 'advanced');
  add_meta_box('routeswaypoints', __('Routes Waypoints', 'routes'),
    'routes_waypoint_inner_edit_form', 'page', 'advanced');
}

/**
 * Prints the inner HTML output for waypoint edit form for a post.
 *
 * @internal
 */
function routes_waypoint_inner_edit_form() {
  global $post_ID, $routesVerifyKey, $routesFileName;
  global $routesBlankLines;
  global $routesEditHeaders, $routesJSForFooter;

  $msg = get_option('routes_last_message_' . $post_ID);
  if ($msg) {
    echo '<div class="updated"><p>' . $msg . '</p></div>' . "\n";
    update_option('routes_last_message_' . $post_ID, '');
  }

  echo "<p>" .  __('Visit the <a href="https://poplarware.com/downloads/routes_install">Routes Plugin</a> page for complete information on the plugin and these settings.', 'routes') . "</p>\n";

  echo "<h4>" . __("Per-post map settings:", 'routes') . "</h4>\n";
  $icons = routes_get_post_icons($post_ID);

  _e("Icon for this post on index maps", 'routes');
  echo routes_icon_select_list('routes_indexicon', $icons[0]);
  echo "<br />\n";

  _e("Icon for start point on single map", 'routes');
  echo routes_icon_select_list('routes_starticon', $icons[1]);
  echo "<br />\n";

  _e("Icon for tagged points on single map", 'routes');
  echo routes_icon_select_list('routes_tagicon', $icons[2]);
  echo "<br />\n";

  _e("Line style", 'routes');
  echo " - ";
  echo routes_line_style_inputs($post_ID);
  echo "<br />\n";

  echo "<h4>" .  __("GPX import on next save:", 'routes') . "</h4>\n";
  echo '<input type="checkbox" name="routes_import_gpx_file" id="routes_import_gpx_file" value="Y" />';
  echo '<label for="routes_import_gpx_file">' . __("Import Waypoints from Uploaded GPX File", 'routes') . '</label> ';

  echo "<br />\n";

  echo '<input type="checkbox" name="routes_import_gpx_in_feet" id="routes_import_gpx_in_feet" value="Y" />';
  echo '<label for="routes_import_gpx_in_feet">' . __("Are elevations in your GPX file in feet? (leave unchecked if they are in meters, which is the standard for GPX files)", 'routes') . '</label> ';

  echo "<br />\n";
  echo '<input type="text" name="routes_import_gpx_skipby" id="routes_import_gpx_skipby" value="1" size="5" />';
  echo '<label for="routes_import_gpx_skipby">' . __("Keep only every N points in GPX (enter 1 to keep all points)", 'routes') . '</label> ';

  echo "<br />\n";

  echo "<ul>\n";

  echo "<h4>" .  __("Route points:", 'routes') . "</h4>\n";

  echo "<p>" .  __("To insert a waypoint higher in the list, enter it in the first blank line and set the Order to a value before or between existing waypoints.", 'routes') . "</p>\n";

  // admin buttons

  echo '<input type="button"' .
    ' name="routes_showhide_ext_btn" id="routes_showhide_ext_btn" value="' .
    __("Show/Hide Extended Info", 'routes') . '" ' .
    ' onclick="routes_showall();return true;"' .
    "/>\n";

  echo '<input type="button"' .
    ' name="routes_delete_all_btn" id="routes_delete_all_btn" value="' .
    __("Mark All Points for Deletion", 'routes') . '" ' .
    ' onclick="routes_mark_all_del();return true;"' .
    "/>\n";

  echo '<input type="button"' .
    ' name="routes_add_more_btn" id="routes_add_more_btn" value="' .
    __("Add More Input Lines", 'routes') . '" ' .
    ' onclick="routes_add_more_lines();return true;"' .
    "/>\n";

  echo "<br />\n";
  echo '<input type="button"' .
    ' name="routes_all_elev_btn" id="routes_all_elev_btn" value="' .
    __("Lookup 100 Elevations", 'routes') . '" ' .
    ' onclick="routes_lookup_100_elev();return true;"' .
    "/>\n";
  echo     __("Starting at row", 'routes') . ' ' .
    '<input type="text" size="5"' .
    ' name="routes_all_elev_start" id="routes_all_elev_start" value="1"' .
    " />\n";

  echo "<br />\n";

  // note: even on page screen, variable is $post_ID here.
  $wpts = routes_read_waypoints($post_ID);

  // print nonce for verification - don't want to save data unless it came
  // from this screen

  echo '<input type="hidden" name="' . $routesVerifyKey .
    '" id="' . $routesVerifyKey . '" value="' .
    wp_create_nonce($routesFileName) . '" />';

  // print current waypoints
  echo "<table>";

  $counter = 0;
  foreach ($wpts as $waypt) {
    if (! ($counter % $routesEditHeaders)) {
      // print header lines every so often...
      $class = 'routes_group_' . $counter;
      $additional = '';
      if ($counter) {
        $additional = ' routes_group_hidden';
      }
      routes_edit_table_header($class, $additional, $counter);
      $class .= $additional;
    }
    $counter++;
    echo '<tr class="' . $class . '">';
    echo '<td><input name="routes_order' . $counter . '" type="text" size="3" value="' . $waypt->waypoint_order . '"/></td>';
    echo '<td><input name="routes_city' . $counter . '" type="text" size="15" value="' . routes_fix_quotes_value($waypt->city) . '"/></td>';
    echo '<td><input name="routes_state' . $counter . '" type="text" size="7" value="' . routes_fix_quotes_value($waypt->state_country) . '"/></td>';
    echo '<td><input name="routes_lat' . $counter . '" type="text" size="6" value="' . $waypt->latitude . '"/></td>';
    echo '<td><input name="routes_long' . $counter . '" type="text" size="6" value="' . $waypt->longitude . '"/></td>';

    // checkbox for delete it
    echo '<td><input class="delcheck" type="checkbox" name="routes_delete' . $counter .
        '"/></td>' . "\n";

    echo "</tr>\n";

    // second row for metadata

     echo '<tr class="routes_extended routes_extended_hidden ' . $class . '">' . "\n";
     echo '<td colspan="5">' . "\n";
    _e('Tag:', 'routes');
    echo '<input name="routes_tag' . $counter . '" type="text" size="35" value="' . routes_fix_quotes_value($waypt->tag) . '"/> ';
    _e('Elevation:', 'routes');
    routes_elev_calc_button($counter);
    echo '<input name="routes_elev' . $counter . '" id="routes_elev' .
      $counter . '" type="text" size="6" value="' . $waypt->elevation . '"/> ';

    echo "</td></tr>\n";

  }

  // print empty lines for new waypoints

  for ($i = 0; $i < $routesBlankLines; $i++) {
    if (! ($counter % $routesEditHeaders)) {
      // print header lines every so often...
      $class = 'routes_group_' . $counter;
      $additional = '';
      if ($counter) {
        $additional = ' routes_group_hidden';
      }
      routes_edit_table_header($class, $additional, $counter);
      $saveclass = $class;
      $class .= $additional;
    }
    $counter++;
    echo '<tr class="' . $class . '">';
    echo '<td><input name="routes_order' . $counter . '" type="text" size="3"/></td>';
    echo '<td><input name="routes_city' . $counter . '" type="text" size="15"/></td>';
    echo '<td><input name="routes_state' . $counter . '" type="text" size="7"/></td>';
    echo '<td><input name="routes_lat' . $counter . '" type="text" size="6"/></td>';
    echo '<td><input name="routes_long' . $counter . '" type="text" size="6"/></td>';

    // checkbox for delete it
    echo '<td><input class="delcheck" type="checkbox" name="routes_delete' . $counter .
        '"/></td>' . "\n";

    echo "</tr>\n";

    // second row for metadata

    echo '<tr class="routes_extended routes_extended_hidden ' . $class . '">' . "\n";
    echo '<td colspan="5">' . "\n";
    _e('Tag:', 'routes');
    echo '<input name="routes_tag' . $counter . '" type="text" size="35" /> ';
    _e('Elevation:', 'routes');
    routes_elev_calc_button($counter);
    echo '<input name="routes_elev' . $counter . '" id="routes_elev' .
      $counter . '" type="text" size="6" /> ';
    echo "</td></tr>\n";

  }

  echo "</table>";

  echo "<h4>" .  __("Points of interest:", 'routes') . "</h4>\n";

  echo '<input type="button"' .
    ' name="routes_add_more_poi_btn" id="routes_add_more_poi_btn" value="' .
    __("Add More Input Lines", 'routes') . '" ' .
    ' onclick="routes_add_more_poi_lines();return true;"' .
    "/>\n";

  $poi = routes_read_poi($post_ID);

  // Print current Points of Interest (POI).
  echo "<table>";

  $counter = 0;
  foreach ($poi as $point) {
    if (! ($counter % $routesEditHeaders)) {
      // print header lines every so often...
      $class = 'routes_poi_group_' . $counter;
      $additional = '';
      if ($counter) {
        $additional = ' routes_group_hidden';
      }
      routes_edit_poi_table_header($class, $additional, $counter);
      $class .= $additional;
    }

    $counter++;
    echo '<tr class="' . $class . '">';
    echo '<td><input name="routes_poi_lat' . $counter . '" type="text" size="6" value="' . $point->latitude . '"/></td>';
    echo '<td><input name="routes_poi_long' . $counter . '" type="text" size="6" value="' . $point->longitude . '"/></td>';
    echo '<td>' . routes_icon_select_list('routes_poi_icon' . $counter, $point->icon_id) . '</td>';

    // checkbox for delete it
    echo '<td><input class="delcheck" type="checkbox" name="routes_poi_delete' . $counter .
        '"/></td>' . "\n";
    echo "</tr>\n";

    // Second row.
    echo '<tr class="' . $class . '">';
    echo '<td colspan="3">' . "\n";
    _e('Text:', 'routes');
    echo '<input name="routes_poi_tag' . $counter . '" type="text" size="50" value="' . routes_fix_quotes_value($point->tag) . '"/> ';
    echo "</td></tr>\n";
  }

  // print empty lines for new POI

  for ($i = 0; $i < $routesBlankLines; $i++) {
    if (! ($counter % $routesEditHeaders)) {
      // print header lines every so often...
      $class = 'routes_poi_group_' . $counter;
      $additional = '';
      if ($counter) {
        $additional = ' routes_group_hidden';
      }
      routes_edit_poi_table_header($class, $additional, $counter);
      $savepoiclass = $class;
      $class .= $additional;
    }
    $counter++;
    echo '<tr class="' . $class . '">';
    echo '<td><input name="routes_poi_lat' . $counter . '" type="text" size="6" /></td>';
    echo '<td><input name="routes_poi_long' . $counter . '" type="text" size="6"/></td>';
    echo '<td>' . routes_icon_select_list('routes_poi_icon' . $counter) . '</td>';

    // checkbox for delete it
    echo '<td><input class="delcheck" type="checkbox" name="routes_poi_delete' . $counter .
        '"/></td>' . "\n";
    echo "</tr>\n";

    // Second row.
    echo '<tr class="' . $class . '">';
    echo '<td colspan="3">' . "\n";
    _e('Text:', 'routes');
    echo '<input name="routes_poi_tag' . $counter . '" id="routes_poi_tag' .
      $counter . '" type="text" size="50" /> ';
    echo "</td></tr>\n";
  }

  echo "</table>";

  // display map for clicking to define points
  routes_map_print($wpts, $poi, 500, 400, "admin", true, '');

  $routesJSForFooter .= "var routesFormLastClass = '$saveclass';\n";
  $routesJSForFooter .= "var routesFormLastPOIClass = '$savepoiclass';\n";
  $routesJSForFooter .= "var routesIconSelect = '" .
    str_replace("'", "\'", routes_icon_select_list('replace_name_here')) . "';";
}


/**
 * Generates and responds to the admin page for plugin option settings.
 *
 * @internal
 */
function routes_options_page() {
  global $routesUseMetric;
  global $routesMakeElev;
  global $routesLookupElev;
  global $routesLeafletElev;
  global $routesElevMarkers;
  global $routesElevWidth, $routesElevHeight;
  global $routesElevLineColor;
  global $routesElevBgColor;
  global $routesElevMarkerColor;
  global $routesElevFillColor;

  /* See if the user has posted us some information */
  if ($_POST['routes_hidden'] == 'Y') {

    // read posted info

    $routesElevWidth = routes_fix_num($_POST['routes_elev_width'], 50, 10000);
    $routesElevHeight = routes_fix_num($_POST['routes_elev_height'], 50, 10000);

    $routesElevLineColor[0] = routes_fix_num($_POST['routes_elev_line_color' . '0' ], 0, 255);
    $routesElevLineColor[1] = routes_fix_num($_POST['routes_elev_line_color' . '1' ], 0, 255);
    $routesElevLineColor[2] = routes_fix_num($_POST['routes_elev_line_color' . '2' ], 0, 255);

    $routesElevBgColor[0] = routes_fix_num($_POST['routes_elev_bg_color' . '0' ], 0, 255);
    $routesElevBgColor[1] = routes_fix_num($_POST['routes_elev_bg_color' . '1' ], 0, 255);
    $routesElevBgColor[2] = routes_fix_num($_POST['routes_elev_bg_color' . '2' ], 0, 255);

    $routesElevFillColor[0] = routes_fix_num($_POST['routes_elev_fill_color' . '0' ], 0, 255);
    $routesElevFillColor[1] = routes_fix_num($_POST['routes_elev_fill_color' . '1' ], 0, 255);
    $routesElevFillColor[2] = routes_fix_num($_POST['routes_elev_fill_color' . '2' ], 0, 255);

    $routesElevMarkerColor[0] = routes_fix_num($_POST['routes_elev_marker_color' . '0' ], 0, 255);
    $routesElevMarkerColor[1] = routes_fix_num($_POST['routes_elev_marker_color' . '1' ], 0, 255);
    $routesElevMarkerColor[2] = routes_fix_num($_POST['routes_elev_marker_color' . '2' ], 0, 255);

    $routesUseMetricVal = routes_num_to_yn(routes_get_yn_option($_POST['routes_metric'], 1));
    $routesMakeElevVal = routes_num_to_yn(routes_get_yn_option($_POST['routes_make_elev'], 1));
    $routesLookupElevVal = routes_num_to_yn(routes_get_yn_option($_POST['routes_lookup_elev'], 1));
    $routesLeafletElevVal = routes_num_to_yn(routes_get_yn_option($_POST['routes_leaflet_elev'], 1));
    $routesElevMarkersVal = routes_num_to_yn(routes_get_yn_option($_POST['routes_elev_markers'], 1));

    $routesIcons = array(
      routes_fix_num($_POST['routes_indexicon'], 1, 10000),
      routes_fix_num($_POST['routes_starticon'], 1, 10000),
      routes_fix_num($_POST['routes_tagicon'], 1, 10000),
    );

    $lineStyle = array($_POST['routes_line_color'], $_POST['routes_line_opacity'], $_POST['routes_line_weight']);

    // save info

    routes_save_post_icons(-1, $routesIcons);
    routes_save_post_line_style(-1, $lineStyle);
    update_option('routes_elev_width', $routesElevWidth);
    update_option('routes_elev_height', $routesElevHeight);
    update_option('routes_use_metric', $routesUseMetricVal);
    update_option('routes_make_elev', $routesMakeElevVal);
    update_option('routes_lookup_elev', $routesLookupElevVal);
    update_option('routes_leaflet_elev', $routesLeafletElevVal);
    update_option('routes_elev_markers', $routesElevMarkersVal);
    update_option('routes_elev_line_color', implode(",", $routesElevLineColor));
    update_option('routes_elev_bg_color', implode(",", $routesElevBgColor));
    update_option('routes_elev_marker_color', implode(",", $routesElevMarkerColor));
    update_option('routes_elev_fill_color', implode(",", $routesElevFillColor));

    // read options again
    routes_setup(1);

    // tell user we updated their info

    echo '<div class="updated"><p><strong>' .
      __('Options saved.', 'routes') .
      "</strong></p></div>";
  }

  /* Now print out the settings form */

  // header/information section

  echo '<div class="wrap">' . "\n";
  echo "<h1>" . __('Routes Settings', 'routes') . "</h1>\n";
  echo "<p>" . __('Visit the <a href="https://poplarware.com/downloads/routes_install">Routes Plugin</a> page for complete information on the plugin and these settings.', 'routes') . "</p>\n";
  echo "<p>" . __('Find RGB color values at these two sites:',
                   'routes') .
    '<a href="https://www.visibone.com/colorlab/">VisiBone</a>, ' .
    '<a href="https://www.lightandcolor.net/colorpick/java-color.html">lightandcolor.net</a>' . "</p>\n";

  echo "<hr />\n";

  // status section

  echo "<h2>" . __("Status", 'routes') . "</h2>\n";

  $reason = routes_is_set_up();
  if (strlen($reason)) {
    echo "<p>";
    _e("Routes is not fully installed and configured. Reason:", 'routes');
    echo "<br>$reason</p>\n";
  } else {
    echo "<p>";
    _e("Routes is installed and configured.", 'routes');
    echo "</p>\n";
  }

  $reason = routes_can_import_gpx();
  if (strlen($reason)) {
    echo "<p>";
    _e("You cannot import GPX files. Reason:", 'routes');
    echo "<br>$reason</p>\n";
  } else {
    echo "<p>";
    _e("Your WordPress/PHP installation should support importing GPX files.", 'routes');
    echo "</p>\n";
  }

  $reason = routes_can_make_elev_graph();
  if (strlen($reason)) {
    echo "<p>";
    _e("You cannot create elevation profiles. Reason:", 'routes');
    echo "<br>$reason</p>\n";
  } else {
    echo "<p>";
    _e("Your WordPress/PHP installation should support creating elevation profiles.", 'routes');
    echo "</p>\n";
  }

  echo "<hr />\n";

  // options form

  echo '<form name="form1" method="post">' . "\n";

  echo '<input type="hidden" name="routes_hidden" value="Y">' . "\n";

  echo '<h2>';
  _e("Icons and line styles", 'routes');
  echo "</h2>\n";

  echo '<p>';
  _e('The following defaults apply to new posts:', 'routes');
  echo "</p>\n";

  $icons = routes_get_post_icons(-1);

  _e("Default icon for posts on index maps", 'routes');
  echo routes_icon_select_list('routes_indexicon', $icons[0]);
  echo "<br />\n";

  _e("Default icon for start point on single maps", 'routes');
  echo routes_icon_select_list('routes_starticon', $icons[1]);
  echo "<br />\n";

  _e("Default icon for tagged points on single maps", 'routes');
  echo routes_icon_select_list('routes_tagicon', $icons[2]);
  echo "<br />\n";

  _e("Default line style", 'routes');
  echo " - ";
  echo routes_line_style_inputs(-1);
  echo "<br />\n";

  echo '<h2>';
  _e("General elevation settings", 'routes');
  echo "</h2>\n";

  echo '<input name="routes_metric" type="checkbox" value="Y" ';
  if ($routesUseMetric) echo "checked";
  echo " />\n";
  _e("Use metric units for elevations", 'routes');
  echo "<br />\n";

  echo '<input name="routes_lookup_elev" type="checkbox" value="Y" ';
  if ($routesLookupElev) echo "checked";
  echo " />\n";
  _e("Look up elevation for each waypoint after clicking map", 'routes');
  echo "<br />\n";

  echo '<h2>';
  _e("Interactive elevation on maps", 'routes');
  echo "</h2>\n";

  echo '<input name="routes_leaflet_elev" type="checkbox" value="Y" ';
  if ($routesLeafletElev) echo "checked";
  echo " />\n";
  _e("Use the interactive Leaflet Elevation plugin on single-post maps", 'routes');
  echo "<br />\n";
  _e('Note: Allow some extra height on your map so the elevation plot will fit!', 'routes');
  echo "<br />\n";

  echo '<h2>';
  _e("Elevation profile images", 'routes');
  echo "</h2>\n";

  echo '<input name="routes_make_elev" type="checkbox" value="Y" ';
  if ($routesMakeElev) echo "checked";
  echo " />\n";
  _e("Create elevation profile image on save", 'routes');
  echo "<br />\n";

  _e("Width of elevation profile image:", 'routes');
  echo ' <input size="7" name="' . 'routes_elev_width' . '" type="text" value="' .
    $routesElevWidth . '"  />' . "\n";
  echo "<br />\n";

  _e("Height of elevation profile image:", 'routes');
  echo ' <input size="7" name="' . 'routes_elev_height' . '" type="text" value="' .
    $routesElevHeight . '"  />' . "\n";
  echo "<br />\n";

  _e("Background color of elevation profile image (R, G, B):", 'routes');
  echo ' <input size="4" name="' . 'routes_elev_bg_color' . '0" type="text" value="' .
    $routesElevBgColor[0] . '"  />' . " ";
  echo ' <input size="4" name="' . 'routes_elev_bg_color' . '1" type="text" value="' .
    $routesElevBgColor[1] . '"  />' . " ";
  echo ' <input size="4" name="' . 'routes_elev_bg_color' . '2" type="text" value="' .
    $routesElevBgColor[2] . '"  />' . " ";
  echo "<br />\n";

  _e("Line/text color of elevation profile image (R, G, B):", 'routes');
  echo '<input size="4" name="' . 'routes_elev_line_color' . '0" type="text" value="' .
    $routesElevLineColor[0] . '"  />' . " ";
  echo '<input size="4" name="' . 'routes_elev_line_color' . '1" type="text" value="' .
    $routesElevLineColor[1] . '"  />' . " ";
  echo '<input size="4" name="' . 'routes_elev_line_color' . '2" type="text" value="' .
    $routesElevLineColor[2] . '"  />' . " ";
  echo "<br />\n";

  _e("Profile fill color of elevation profile image (R, G, B):", 'routes');
  echo '<input size="4" name="' . 'routes_elev_fill_color' . '0" type="text" value="' .
    $routesElevFillColor[0] . '"  />' . " ";
  echo '<input size="4" name="' . 'routes_elev_fill_color' . '1" type="text" value="' .
    $routesElevFillColor[1] . '"  />' . " ";
  echo '<input size="4" name="' . 'routes_elev_fill_color' . '2" type="text" value="' .
    $routesElevFillColor[2] . '"  />' . " ";
  echo "<br />\n";

  echo '<input name="routes_elev_markers" type="checkbox" value="Y" ';
  if ($routesElevMarkers) echo "checked";
  echo " />\n";
  _e("Add markers to elevation profiles", 'routes');
  echo "<br />\n";

  _e("Marker color of elevation profile image (R, G, B):", 'routes');
  echo ' <input size="4" name="' . 'routes_elev_marker_color' . '0" type="text" value="' .
    $routesElevMarkerColor[0] . '"  />' . " ";
  echo ' <input size="4" name="' . 'routes_elev_marker_color' . '1" type="text" value="' .
    $routesElevMarkerColor[1] . '"  />' . " ";
  echo ' <input size="4" name="' . 'routes_elev_marker_color' . '2" type="text" value="' .
    $routesElevMarkerColor[2] . '"  />' . " ";
  echo "<br />\n";


  echo '<p class="submit">' . "\n";

  echo '<input type="submit" name="Submit" value="' .
     __('Update Options', 'routes') . '" />' . "\n";

  echo "</p></form></div>\n";
}

/**
 * Updates the Routes information when a post or page is saved.
 *
 * @internal
 */
function routes_post_update($post_id) {
  global $routesVerifyKey, $routesFileName;
  global $wpdb;
  global $routesWaypointTable;
  global $routesPOITable;
  global $routesMakeElev;

  static $called_posts = array();

  // verify this came from the our screen and with proper authorization
  if (!wp_verify_nonce($_POST[$routesVerifyKey], $routesFileName)) {
    return $post_id;
  }

  // Load the post and verify this is being called for editing the post we
  // think it is. Reason: this could be called for an attachment we are saving,
  // such as the elevation graph.
  $post = get_post($post_id);
  if ($post->post_type != $_POST['post_type']) {
    return $post_id;
  }
  if ($_POST['post_ID'] != $post_id) {
    return $post_id;
  }

  if ('page' == $_POST['post_type']) {
    if (!current_user_can('edit_page', $post_id)) {
      return $post_id;
    }
  } else if ('post' == $_POST['post_type']) {
    if (!current_user_can('edit_post', $post_id)) {
      return $post_id;
    }
  }
  else {
    // it was an attachment or some other kind of post
    return $post_id;
  }

  // Make sure we haven't already done our thing for this post. Reason: WP,
  // in its wisdom, has this function called  _fix_attachment_links() that is
  // called from wp_update_post(), which then calls wp_update_post() again,
  // triggering a call to this function. But if we have already processed
  // the waypoint information in $_POST, there is no reason to do it again.
  // Not only that, but if we did do it again, GPX import would be screwed up,
  // because we would have deleted the GPX file.
  if (isset($called_posts[$post_id])) {
    return;
  }
  $called_posts[$post_id] = 1;

  $msg = get_option('routes_last_message_' . $post_id);
  $msg .= '<br />' . __('Routes messages for post', 'routes') . ' ' . $post_id;

  // Delete old information.
  routes_post_delete($post_id);

  // Update the icon and line style information.
  $info = array($_POST['routes_indexicon'], $_POST['routes_starticon'], $_POST['routes_tagicon']);
  routes_save_post_icons($post_id, $info);
  $info = array($_POST['routes_line_color'], $_POST['routes_line_opacity'], $_POST['routes_line_weight']);
  routes_save_post_line_style($post_id, $info);

  // Update the waypoint information.
  // read info from form post and insert into db

  $i = 1;
  $highest_order = 0;
  $numpts = 0;

  while (strlen($_POST[ 'routes_order' . $i  ]) > 0 ||
         strlen($_POST[ 'routes_lat' . $i  ]) > 0 ||
         strlen($_POST[ 'routes_long' . $i  ]) > 0 ||
         strlen($_POST[ 'routes_tag' . $i  ]) > 0 ||
         strlen($_POST[ 'routes_elev' . $i  ]) > 0) {
    if (!$_POST[ 'routes_delete' . $i ]) {

      // user didn't check DELETE on this point, so save info

      if ((int) $_POST[ 'routes_order' . $i ] > $highest_order) {
        $highest_order = (int) $_POST[ 'routes_order' . $i ];
      }

      $numpts++;

      $query = "INSERT INTO $routesWaypointTable (post_id, waypoint_order, city, state_country, latitude, longitude, elevation, tag) VALUES ($post_id, ";

      // Question: does this need to be better sanitized? $_POST seems to be
      // escaping the quotes correctly...
      $query .= "'" . routes_numpart($_POST[ 'routes_order' . $i ])
        . "', ";
      $query .= "'" . $_POST[ 'routes_city' . $i ] . "', ";
      $query .= "'" . $_POST[ 'routes_state' . $i ] . "', ";
      $query .= "'" . routes_numpart($_POST[ 'routes_lat' . $i ])
        . "', ";
      $query .= "'" . routes_numpart($_POST[ 'routes_long' . $i ])
        . "', ";
      $query .= "'" . routes_numpart($_POST[ 'routes_elev' . $i ])
        . "', ";
      $query .= "'" . $_POST[ 'routes_tag' . $i ]. "') ";

      $wpdb->query($query);
    }
    $i++;

  }

  if ($numpts) {
    $msg .= '<br />' . __('Saved points from screen', 'routes') . ' ' . $numpts;
  }

  // If box was checked, import from attached GPX file

  if ($_POST[ 'routes_import_gpx_file' ] == "Y") {
    routes_import_gpx($post_id, $highest_order, $msg);
  }

  // make elevation graph if we are supposed to

  if ($routesMakeElev) {
    routes_make_elev_graph($post_id, $msg);
  }

  // Points of Interest

  $i = 1;
  $numpts = 0;

  while (strlen($_POST[ 'routes_poi_lat' . $i  ]) > 0 ||
    strlen($_POST[ 'routes_poi_long' . $i  ]) > 0 ||
    strlen($_POST[ 'routes_poi_tag' . $i  ]) > 0) {

    if (!$_POST[ 'routes_poi_delete' . $i ]) {
      // user didn't check DELETE on this point, so save info
      $numpts++;

      // Question: does this need to be better sanitized? $_POST seems to be
      // escaping the quotes correctly...
      $query = "INSERT INTO $routesPOITable (post_id, latitude, longitude, icon_id, tag) VALUES ($post_id, ";

      $query .= "'" . routes_numpart($_POST[ 'routes_poi_lat' . $i ])
        . "', ";
      $query .= "'" . routes_numpart($_POST[ 'routes_poi_long' . $i ])
        . "', ";
      $query .= "'" . routes_numpart($_POST[ 'routes_poi_icon' . $i ])
        . "', ";
      $query .= "'" . $_POST[ 'routes_poi_tag' . $i ]. "') ";

      $wpdb->query($query);
    }
    $i++;

  }

  if ($numpts) {
    $msg .= '<br />' . __('Saved points of interest from screen', 'routes') . ' ' . $numpts;
  }

  update_option('routes_last_message_' . $post_id, $msg);

  return $post_id;
}

/**
 * Deletes the Routes information when a post/page is deleted.
 *
 * @internal
 */
function routes_post_delete($post_id) {
  global $wpdb;
  global $routesWaypointTable;
  global $routesPOITable;
  global $routesPostIconTable;
  global $routesPostLineStyleTable;

  $tables = array($routesWaypointTable, $routesPOITable, $routesPostIconTable, $routesPostLineStyleTable);
  $id = intval($post_id);

  foreach ($tables as $table) {
    $query = "DELETE FROM $table WHERE post_id=$post_id";
    $wpdb->query($query);
  }

  return $post_id;
}

// Generates the admin page for managing icons,
// and responds when the user updates the icons.
function routes_icons_page() {
  global $routesIconTable;
  global $wpdb;

  /* See if the user has posted us some information */
  if ($_POST['routes_hidden'] == 'Y') {
    $nrows = intval($_POST['routes_icon_num_rows']);
    for ($i = 0; $i < $nrows; $i++) {
      $id = intval($_POST['routes_icon_id' . $i]);

      // See if they requested a delete
      if (routes_get_yn_option($_POST['routes_icon_delete' . $i], 1)) {
        if ($id > 0) {
          $qry = "DELETE FROM $routesIconTable WHERE icon_id=$id";
          $wpdb->query($qry);
        }
        continue;
      }

      // Read the information and save it.

      $aname = $wpdb->escape($_POST['routes_icon_admin_name' . $i]);
      $lname = $wpdb->escape($_POST['routes_icon_legend_name' . $i]);
      $w = intval($_POST['routes_icon_size_x' . $i]);
      $h = intval($_POST['routes_icon_size_y' . $i]);
      $sw = intval($_POST['routes_shadow_size_x' . $i]);
      $sh = intval($_POST['routes_shadow_size_y' . $i]);
      $x = intval($_POST['routes_icon_anchor_x' . $i]);
      $y = intval($_POST['routes_icon_anchor_y' . $i]);
      $sx = intval($_POST['routes_shadow_anchor_x' . $i]);
      $sy = intval($_POST['routes_shadow_anchor_y' . $i]);
      $f = $wpdb->escape(routes_read_icon_file('routes_icon_file' . $i));
      $sf = $wpdb->escape(routes_read_icon_file('routes_shadow_file' . $i));

      if (!$aname) {
        continue;
      }

      if ($id > 0) {
        $qry = "UPDATE $routesIconTable SET admin_name='$aname', legend_name='$lname', icon_image='$f', shadow_image='$sf', size_x=$w, size_y=$h, shadow_size_x=$sw, shadow_size_y=$sh, anchor_x=$x, anchor_y=$y, shadow_anchor_x=$sx, shadow_anchor_y=$sy WHERE icon_id=$id";
        $wpdb->query($qry);
      }
      else {
        $qry = "INSERT INTO $routesIconTable (admin_name, legend_name, icon_image, shadow_image, size_x, size_y, shadow_size_x, shadow_size_y, anchor_x, anchor_y, shadow_anchor_x, shadow_anchor_y) VALUES ('$aname', '$lname', '$f', '$sf', $w, $h, $sw, $sh, $x, $y, $sx, $sy)";
        $wpdb->query($qry);
      }
    }

    // Tell the user we saved their information.
    echo '<div class="updated"><p><strong>' .
      __('Options saved.', 'routes') .
      "</strong></p></div>";
  }

  echo '<div class="wrap">' . "\n";
  echo "<h2>" . __('Routes Icons', 'routes') . "</h2>\n";

  echo '<form name="form1" enctype="multipart/form-data" method="post">' . "\n";
  echo '<input type="hidden" name="routes_hidden" value="Y">' . "\n";

  echo '<p>' . __('All fields are required except Legend Name, and you can also omit the Shadow completely. Leave the Legend Name blank to omit an icon from the legend.', 'routes') . '</p>';
  echo '<p>' . __('Note that GIF images may not work as file uploads. If you have problems, try entering them as a URL, which seems to work better.', 'routes') . '</p>';
  echo '<p>' . __('If you delete an icon that was in use, you may have problems viewing maps!', 'routes') . '</p>';

  // List the icons.
  $qry = "SELECT * FROM $routesIconTable ORDER BY admin_name";
  $results = $wpdb->get_results($qry);

  $style = ' style="border:1px solid black;vertical-align: top; border-collapse: collapse;"';
  echo '<table' . $style . '">';
  echo "<tr><th" . $style . ">" . __('Icon and Shadow', 'routes') . '</th><th' . $style . '>' .
    __('Administrative Name', 'routes') . '</th><th' . $style . '>' .
    __('Legend Name', 'routes') . '</th><th' . $style . '>' .
    __('Files', 'routes') . '</th><th' . $style . '>' .
    __('Details', 'routes') . "</th><th></th></tr>\n";

  $i = 0;
  foreach ($results as $icon) {
    echo '<tr' . $style . '>';
    echo '<td' . $style . '><img src="' . $icon->icon_image . '" alt="' .
      __('Icon image', 'routes') . '" />' . '<br />';
    if ($icon->shadow_image) {
      echo '<img src="' . $icon->shadow_image . '" alt="' .
      __('Shadow image', 'routes') . '" />';
    }
    echo '</td>';
    routes_icon_edit_row($icon, $i);
    $i++;
    echo '</tr>';
  }

  echo '<tr><td' . $style . '>' . __('Add new', 'routes') . '</td>';
  echo routes_icon_edit_row(NULL, $i);
  $i++;
  echo '</tr>';

  echo '</table>' . "\n";

  echo '<input type="hidden" value="' . $i . '" name="routes_icon_num_rows" />' . "\n";
  echo '<p class="submit">' . "\n";

  echo '<input type="submit" name="Submit" value="' .
     __('Update Icons', 'routes') . '" />' . "\n";

  echo "</p></form></div>\n";
}

/**
 * Builds an HTML select list for icons.
 *
 * @internal
 */
function routes_icon_select_list($name, $selected_id = -1) {
  global $wpdb;
  global $routesIconTable;

  $qry = "SELECT icon_id, admin_name FROM $routesIconTable";
  $results = $wpdb->get_results($qry);
  $out = '<select name="' . $name . '"> ';
  foreach ($results as $icon) {
    $out .= '<option value="' . $icon->icon_id . '"';
    if ($icon->icon_id == $selected_id) {
      $out .= " selected";
    }
    $out .= '>' . $icon->admin_name . '</option> ';
  }

  $out .= '</select>';
  return $out;
}

/**
 * Returns HTML for inputs for specifying line style.
 *
 * @internal
 */
function routes_line_style_inputs($post_id) {
  $curr = routes_get_post_line_style($post_id);
  $out = '';

  $out .= __('Color:', 'routes') . ' #<input type="text" value="' .
    $curr[0] . '" name="routes_line_color" size="6" /> ';
  $out .= __('Opacity (0-1):', 'routes') . ' #<input type="text" value="' .
    $curr[1] . '" name="routes_line_opacity" size="5" /> ';
  $out .= __('Weight (pixels):', 'routes') . ' #<input type="text" value="' .
    $curr[2] . '" name="routes_line_weight" size="3" /> ';

  return $out;
}

/**
 * Saves icons to use for the given post.
 *
 * @param int $id
 *   ID to save icons for.
 * @param array $values
 *   Icons to save.
 *
 * @internal
 */
function routes_save_post_icons($id, $values) {
  global $wpdb;
  global $routesPostIconTable;

  // Sanitize.
  $id = intval($id);

  // Remove old values, if any.
  $qry = "DELETE FROM $routesPostIconTable WHERE post_id = $id";
  $wpdb->query($qry);

  // Insert new values.
  $qrystart = "INSERT INTO $routesPostIconTable (post_id, icon_id, icon_type) VALUES ";
  for ($i = 0; $i < 3; $i++) {
    $num = $i + 1;
    $qry = $qrystart . '(' . $id . ', ' . intval($values[$i]) . ', ' . $num . ')';
    $wpdb->query($qry);
  }
}

/**
 * Saves line style to use for the given post.
 *
 * @param int $id
 *   ID to save line style for.
 * @param array $values
 *   Line style to save.
 *
 * @internal
 */
function routes_save_post_line_style($id, $values) {
  global $wpdb;
  global $routesPostLineStyleTable;

  // Sanitize.
  $id = intval($id);
  $color = preg_replace("/[^a-zA-Z0-9]/", "", $values[0]);
  $opacity = routes_fix_num(floatval($values[1]), 0, 1);
  $weight = routes_fix_num(intval($values[2]), 1, 10000);

  // Remove old values, if any.
  $qry = "DELETE FROM $routesPostLineStyleTable WHERE post_id = $id";
  $wpdb->query($qry);

  // Insert new values.
  $qry = "INSERT INTO $routesPostLineStyleTable (post_id, color, opacity, weight) VALUES ($id, '$color', '$opacity', $weight)";
  $wpdb->query($qry);
}

/**
 * Prints out an icon editing row for the icon screen.
 *
 * @param object $icon
 *   Current icon information (if any).
 * @param int $index
 *   Index for this row.
 *
 * @internal
 */
function routes_icon_edit_row($icon, $index) {

  $style = ' style="border:1px solid black;vertical-align: top;"';

  // Admin name.
  echo '<td' . $style . '><input type="text" name="routes_icon_admin_name' . $index . '"';
  if (isset($icon->admin_name)) {
    echo ' value="' . routes_fix_quotes_value($icon->admin_name) . '"';
  }
  echo " /></td>";

  // Legend.
  echo '<td' . $style . '><input type="text" name="routes_icon_legend_name' . $index . '"';
  if (isset($icon->legend_name)) {
    echo ' value="' . routes_fix_quotes_value($icon->legend_name) . '"';
  }
  echo " /></td>";

  // Files.
  echo '<td' . $style . '>';
  $iconimg = isset($icon->icon_image) ? $icon->icon_image : '';
  $shadowimg = isset($icon->shadow_image) ? $icon->shadow_image : '';
  echo routes_file_or_url_input($iconimg, __('Icon', 'routes'), 'routes_icon_file' . $index);
  echo routes_file_or_url_input($shadowimg, __('Shadow', 'routes'), 'routes_shadow_file' . $index);

  // Everything else.
  echo '<td' . $style . '>';

  echo '<input type="hidden" name="routes_icon_id' . $index . '" value="';
  if (isset($icon->icon_id)) {
    echo $icon->icon_id;
  }
  else {
    echo '-1';
  }
  echo '" />';

  echo __('Icon width', 'routes') .
    '<input  size="5" type="text" name="routes_icon_size_x' . $index . '"';
  if (isset($icon->size_x)) {
    echo ' value="' . $icon->size_x . '"';
  }
  echo " /><br />";
  echo __('Icon height', 'routes') .
    '<input size="5"  type="text" name="routes_icon_size_y' . $index . '"';
  if (isset($icon->size_y)) {
    echo ' value="' . $icon->size_y . '"';
  }
  echo " /><br />";

  echo __('Shadow width', 'routes') .
    '<input type="text"  size="5" name="routes_shadow_size_x' . $index . '"';
  if (isset($icon->shadow_size_x)) {
    echo ' value="' . $icon->shadow_size_x . '"';
  }
  echo " /><br />";
  echo __('Shadow height', 'routes') .
    '<input type="text"  size="5" name="routes_shadow_size_y' . $index . '"';
  if (isset($icon->shadow_size_y)) {
    echo ' value="' . $icon->shadow_size_y . '"';
  }
  echo " /><br />";

  echo __('Icon anchor position x', 'routes') .
    '<input size="5" type="text"  size="5" name="routes_icon_anchor_x' . $index . '"';
  if (isset($icon->anchor_x)) {
    echo ' value="' . $icon->anchor_x . '"';
  }
  echo " /><br />";
  echo __('Icon anchor position y', 'routes') .
    '<input type="text"  size="5" name="routes_icon_anchor_y' . $index . '"';
  if (isset($icon->anchor_y)) {
    echo ' value="' . $icon->anchor_y . '"';
  }
  echo " /><br />";

  echo __('Shadow anchor position x', 'routes') .
    '<input type="text"  size="5" name="routes_shadow_anchor_x' . $index . '"';
  if (isset($icon->shadow_anchor_x)) {
    echo ' value="' . $icon->shadow_anchor_x . '"';
  }
  echo " /><br />";
  echo __('Shadow anchor position y', 'routes') .
    '<input type="text"  size="5" name="routes_shadow_anchor_y' . $index . '"';
  if (isset($icon->shadow_anchor_y)) {
    echo ' value="' . $icon->shadow_anchor_y . '"';
  }
  echo " />";

  echo "</td>";

  // Delete checkbox
  echo '<td' . $style . '><input type="checkbox" name="routes_icon_delete' . $index . '" value="Y" />' . __('Delete this icon', 'routes') . '</td>';
}

/**
 * Prints out HTML inputs for an image or URL.
 *
 * @param string $fname
 *   File name for the icon.
 * @param string $title
 *   Title for the HTML area.
 * @param string $inputname
 *   HTML name for the input.
 *
 * @internal
 */
function routes_file_or_url_input($fname, $title, $inputname) {

  echo '<strong>' . $title . "</strong><br />";

  echo __('URL:', 'routes') . '<input type="text" name="' . $inputname . '_replace_url" size="30"  value="' .
      routes_fix_quotes_value($fname) . '" /><br />';
  echo __('Replace with file:', 'routes') . '<input type="file" name="' . $inputname . '_replace_file" /><br />';
}

/**
 * Makes an elevation graph for the given post and saves as attachment.
 *
 * @param int $post_id
 *   ID of the post or page.
 * @param string $msg
 *   In-progress message to add to.
 */
function routes_make_elev_graph ($post_id, &$msg) {
  global $routesUseMetric;
  global $routesElevWidth, $routesElevHeight;
  global $routesElevLineColor, $routesElevBgColor, $routesElevFillColor;
  global $routesElevMarkers, $routesElevMarkerColor;
  global $wpdb;

  if (strlen(routes_can_make_elev_graph())) {
    // can't do elevation graphs
    return;
  }

  // figure out where to upload files

  $uploads = routes_graph_dir();

  // create blank image object

  $width = $routesElevWidth;
  $height = $routesElevHeight;

  $img = imagecreate($width, $height);
  if ($img === false) {
    return;
  }

  $lineColor = imagecolorallocate($img,
                                   $routesElevLineColor[0],
                                   $routesElevLineColor[1],
                                   $routesElevLineColor[2]);
  $bgColor = imagecolorallocate($img,
                                 $routesElevBgColor[0],
                                 $routesElevBgColor[1],
                                 $routesElevBgColor[2]);
  $fillColor = imagecolorallocate($img,
                                   $routesElevFillColor[0],
                                   $routesElevFillColor[1],
                                   $routesElevFillColor[2]);

  if ($routesElevMarkers) {
    $markerColor = imagecolorallocate($img,
                                   $routesElevMarkerColor[0],
                                   $routesElevMarkerColor[1],
                                   $routesElevMarkerColor[2]);
  }
  else {
    $markerColor = $bgColor;
  }

  if ($lineColor === false || $bgColor === false || $fillColor === false || $markerColor == FALSE) {
    return;
  }

  imagefill($img, 0, 0, $bgColor);

  // process waypoint data -- find x,y for each elevation

  $wpts = routes_read_waypoints($post_id);
  if (!count($wpts)) {
    return;
  }

  $minElev = 99999;
  $maxElev = -99999;
  $x = array();
  $y = array();

  $markerx = array();
  $markery = array();

  // distance conversion factors
  $piconv = pi() / 180.0;
  $rad = 3956;
  if ($routesUseMetric) {
    $rad = 6367;
  }

  $prevlat = $wpts[0]->latitude * $piconv;
  $prevlong = $wpts[0]->longitude * $piconv ;
  $dist = 0;
  $thisx = 0;
  $thisy = 0;

  foreach ($wpts as $waypoint) {
    // add on distance from prev point to this one

    // Haversine formula for distance
    // https://www.faqs.org/faqs/geography/infosystems-faq/
    $lat = $waypoint->latitude * $piconv;
    $long = $waypoint->longitude * $piconv;

    $dlon = 0.5 * ($long - $prevlong);
    $dlat = 0.5 * ($lat - $prevlat);
    $a = sin($dlat) * sin($dlat) + cos($lat) * cos($lat) *
      sin($dlon) * sin($dlon);
    $a = routes_fix_num($a, 0, 1);
    $a = 2.0 * asin(sqrt($a));
    $dist += $rad * $a;

    if (strlen(trim($waypoint->elevation)) >= 1 &&
        ($waypoint->elevation > -200)) {
      $thisy = $waypoint->elevation;
      $thisx = $dist;
      $y[] = $thisy;
      $x[] = $thisx;

      if ($waypoint->elevation > $maxElev) {
        $maxElev = $waypoint->elevation;
      }
      if ($waypoint->elevation < $minElev) {
        $minElev = $waypoint->elevation;
      }
    }

    if (strlen(trim($waypoint->tag)) > 0) {
      $markerx[] = $thisx;
      $markery[] = $thisy;
    }

    $prevlat = $lat;
    $prevlong = $long;
  }

  if (count($x) < 2 || $dist <= 0 ||
      $maxElev < $minElev) {
    // no points with elevation found - must have 2 to plot
    return;
  }

  // round limiting elevations to whole numbers
  $maxElev = ceil($maxElev);
  $minElev = floor($minElev);

  // calculate plotting scales and offsets
  // Remember, y min is at top of image!

  $fontNum = 2;
  $fh = imagefontheight($fontNum);
  $fw = imagefontwidth($fontNum);
  $chars = 1;
  if (strlen($maxElev) > $chars) {
    $chars = strlen($maxElev);
  }
  if (strlen($minElev) > $chars) {
    $chars = strlen($minElev);
  }
  if (strlen($dist) > $chars) {
    $chars = strlen($maxElev);
  }
  $margin = $fw * $chars;
  $extra = 5;

  $minx = $margin + $extra;
  $maxx = $width - $margin;
  $scalex = ($maxx - $minx) / $dist;

  $miny = $margin;
  $maxy = $height - $margin - $extra;
  $scaley = 0;

  if ($maxElev > $minElev) {
    $scaley = ($maxy - $miny) / ($maxElev - $minElev);
  } else {
    // max/min same elevation
    $miny = 0.5 * ($miny + $maxy);
  }

  // plot points as a filled polygon

  $pts = array();

  $pts[] = $minx + $scalex * ($x[0]);
  $pts[] = $maxy;

  for ($i = 0; $i < count($x); $i++) {
    $thisx = $minx + $scalex * ($x[$i]);
    $thisy = $miny + $scaley * ($maxElev - $y[$i]);
    $pts[] = $thisx;
    $pts[] = $thisy;
  }

  $pts[] = $thisx;
  $pts[] = $maxy;
  imagefilledpolygon($img, $pts, count($pts) / 2, $fillColor);

  if ($routesElevMarkers && count($markerx)) {
    for ($i = 0; $i < count($markerx); $i++) {
      $thisx = $minx + $scalex * ($markerx[$i]);
      $thisy = $miny + $scaley * ($maxElev - $markery[$i]);

      $pts = array($thisx - 5, $thisy - 5, $thisx - 5, $thisy + 5,
        $thisx + 5, $thisy + 5, $thisx + 5, $thisy - 5);
      imagefilledpolygon($img, $pts, 4, $markerColor);
    }
  }
  else {
    $msg .= "<br />No markers because boolean was $routesElevMarkers and count was " . count($markerx);
  }


  // add axes lines and labels (labels only at min/max values)

  imageline($img, $minx, $maxy, $maxx, $maxy, $lineColor);
  imageline($img, $minx, $miny, $minx, $maxy, $lineColor);

  imagestring($img, $fontNum, $extra, $miny - 0.5 * $fh, $maxElev, $lineColor);
  imagestring($img, $fontNum, $extra, $maxy - 0.5 * $fh, $minElev, $lineColor);

  // figure out where to put image and delete if it was there

  $fname = routes_elev_graph_filename($post_id);

  // see if the attachment already existed, and delete if so

  $att_id = (int) $wpdb->get_var("SELECT post_id FROM " . $wpdb->postmeta .
    " WHERE meta_key='_wp_attached_file' AND meta_value LIKE '%" . $wpdb->escape($fname) . "%'");
  if ($att_id) {
    wp_delete_attachment($att_id);
    $msg .= '<br />' . __('Deleted previous elevation graph', 'routes') ;
  }

  // create image file

  $new_file = $uploads[ 'path' ] . "/" . $fname;
  $ok = imagepng($img, $new_file);
  imagedestroy($img);
  if (!$ok) {
    return;
  }

  // set file permissions

  $stat = stat(dirname($new_file));
  $perms = $stat['mode'] & 0000666;
  @ chmod($new_file, $perms);

  // add attachment meta-data

  $url = $uploads[ 'url' ] . "/" . $fname;

  $title = __("Routes Elevation Plot");
  $attachment = array(
                      'post_title' => $title,
                      'post_content' => $title,
                      'post_type' => 'attachment',
                      'post_parent' => $post_id,
                      'post_mime_type' => 'image/png',
                      'guid' => $url
                     );

  $id = wp_insert_attachment($attachment, $new_file, $post_id);

  wp_update_attachment_metadata($id,
                                 wp_generate_attachment_metadata($id, $new_file));
    $msg .= '<br />' . __('Created new elevation graph', 'routes');
}

/**
 * Imports a GPX file attached to a post, if there is one.
 *
 * @param int $post_id
 *   ID of post to import for.
 * @param int $highest_order
 *   Starts with order numbering $highest_order + 10.
 * @param string $msg
 *   In-progress message for editing screen after save.
 */
function routes_import_gpx($post_id, &$highest_order, &$msg) {

  global $wpdb;
  global $routesGPXMimeType;
  global $routesUseMetric;
  global $routesWaypointTable;

  if ($why = strlen(routes_can_import_gpx()) != 0) {
    $msg .= '<br />' . __("You cannot import GPX files. Reason:", 'routes') . '<br />&nbsp;&nbsp; '  . $why;
    return;
  }

  $ok = FALSE;
  $num_pts = 0;

  // find attachment

  $post_id = intval($post_id);
  $query = "SELECT * FROM $wpdb->posts WHERE post_type='attachment' AND post_parent='$post_id' AND post_mime_type='$routesGPXMimeType'";
  $info = $wpdb->get_row($query);

  if (!$info || !$info->ID) {
    $msg .= '<br />' .  __('No GPX attachment found for post', 'routes') . ' ' . $post_id;
    return;
  }

  // open attachment in XML parser

  $file = get_attached_file($info->ID);
  if (!$file || !file_exists($file)) {
    $msg .= '<br />' . __('GPX file could not be opened', 'routes');
    return;
  }

  $dom = 0;
  $dom = new DOMDocument();
  $ok = $dom->load($file);

  if (!$ok) {
    $msg .= '<br />' . __('Could not parse GPX file', 'routes');
    return;
  }

  // figure out conversion factor, if necessary

  $conv = 1.0;
  $importFeet = ($_POST[ 'routes_import_gpx_in_feet' ] == 'Y');
  if ($importFeet && $routesUseMetric) {
    // need to convert from feet to meters
    $conv = 0.3048;
  } else if (!$importFeet && !$routesUseMetric) {
    // need to convert from meters to feet
    $conv = 3.280833;
  }

  // import waypoints from attachment

  $skipby = intval($_POST[ 'routes_import_gpx_skipby' ]);
  if ($skipby < 2) {
    $skipby = 1;
  }
  // Start skip count at minus one so the first point is not skipped.
  $skipcount = -1;

  $tracks = array();
  $root = $dom->documentElement;
  $tracks = $root->getElementsByTagName('trk');

  foreach ($tracks as $track) {
    $segs = array();
    $segs = $track->getElementsByTagName('trkseg');

    foreach ($segs as $seg) {
      $pts = array();
      $pts = $seg->getElementsByTagName('trkpt');

      foreach ($pts as $pt) {

        // see if we should skip this point
        $skipcount++;
        if ($skipcount % $skipby) {
          // we are skipping this point
          continue;
        }

        $lat = 0;
        $long = 0;
        $elev = 0;

        $lat = routes_numpart($pt->getAttribute('lat'));
        $long = routes_numpart($pt->getAttribute('lon'));
        $elevs = $pt->getElementsByTagName('ele');
        if ($first = $elevs->item(0)) {
          $elev = $conv * routes_numpart($first->textContent);
        }

        // OK, we have a waypoint, add it

        $highest_order += 10;
        $num_pts++;

        $query = "INSERT INTO $routesWaypointTable (post_id, waypoint_order, latitude, longitude, elevation) VALUES ('$post_id', '$highest_order', '$lat', '$long', '$elev')";

        $wpdb->query($query);
      }
    }
  }

  if ($num_pts) {
    // Remove attachment if all went well.
    wp_delete_attachment($info->ID);
    $msg .= '<br />' . __('Successfully imported and deleted GPX file', 'routes');
  }
  else {
    $msg .= '<br />' . __('No points found in GPX file', 'routes');
  }
}

/**
 * Checks to see if XML parsing functions are there for GPX import.
 *
 * @return string
 *   Empty string if import should work, or a message explaining why if not.
 */
function routes_can_import_gpx() {

  // check for XML library

  if (!routes_php5_xml_lib_exists()) {
    return __("The XML DOM Parsing library is not installed on your web site host", 'routes');
  }

  return '';
}

/**
 * Checks to see if elevation graphs can be made or not.
 *
 * @return string
 *   Empty string if they can, or a message explaining why if not.
 *
 * @internal
 */
function routes_can_make_elev_graph() {

  if (!current_user_can('upload_files')) {
    return __("You do not have permission to upload files", 'routes');
  }

  $uploads = routes_graph_dir();
  if ($uploads[ 'error' ]) {
    return __("Your uploads directory is read-only, so a graph cannot be saved there", 'routes');
  }

  // check for GD library

  if (!function_exists('imagecreate') ||
      !function_exists('imagecolorallocate') ||
      !function_exists('imagefontheight') ||
      !function_exists('imagefontwidth') ||
      !function_exists('imagefilledpolygon') ||
      !function_exists('imageline') ||
      !function_exists('imagestring') ||
      !function_exists('imagepng') ||
      !function_exists('imagedestroy')) {
    return __("The GD library is not installed on your web site host", 'routes');
  }

  // everything is OK
  return '';
}

/**
 * Checks to see if Routes is installed correctly.
 *
 * @return string
 *   Empty string if it is set up fine, or a message explaining why if not.
 */
function routes_is_set_up() {
  global $routesWaypointTable;
  global $wpdb;
  global $routesDBVersion;
  global $routesPOITable;
  global $routesIconTable;
  global $routesPostIconTable;
  global $routesPostLineStyleTable;

  $ret = '';

  // Try to re-install the tables, just in case.
  routes_install_tables();

  $currver = get_option("routes_db_version");
  if ($currver != $routesDBVersion) {
    $ret .= __("You have uploaded a new Routes version, and the Routes database needs to be updated. Deactivate the Routes plugin and reactivate it to update the database.", 'routes') . "<br />\n";
  }

  // Check for the existence of Routes tables.
  $tables = array($routesWaypointTable, $routesPOITable, $routesIconTable, $routesPostIconTable, $routesPostLineStyleTable);
  foreach ($tables as $table) {
    $qry = "SHOW TABLES LIKE '$table'";
    $tbl = $wpdb->get_results($qry);
    if (!count($tbl)) {
      $ret .= __("One or more database tables for Routes do not exist. Probably when you installed Routes, you did not have permission to create new tables in your Wordpress database. Contact your web host to correct this problem, then deactive Routes, activate Routes again, and check back to see if the problem is fixed.", 'routes') . "<br />\n";
      break;
    }
  }

  $uploads = routes_graph_dir();
  if ($uploads[ 'error' ]) {
    $ret .= __("You cannot fully use the custom icons feature of Routes, because Routes cannot save icon files in your uploads directory. Please ensure that your uploads directory is not read-only.", 'routes');
  }

  // Check for at least one icon being present.
  $qry = "SELECT * FROM $routesIconTable ORDER BY admin_name";
  $results = $wpdb->get_results($qry);
  if (!count($results)) {
    $ret .= __("There are no icons, so map markers will not work.", 'routes');
  }

  return $ret;
}

/**
 * Reads information from $_POST for an icon file for the icon page.
 *
 * @param string $prefix
 *   Prefix for this post for the icon information.
 *
 * @return string
 *   URL of the icon file.
 *
 * @internal
 */
function routes_read_icon_file($prefix) {
  $cur = $_POST[$prefix . '_replace_url'];

  if ($_FILES[$prefix . '_replace_file']) {
    $overrides = array( 'test_form' => false);
    $file = wp_handle_upload($_FILES[$prefix . '_replace_file'], $overrides);
    if ($file['url']) {
      return $file['url'];
    }
  }

  return $cur;
}

/**
 * Prints table header line for the edit screen, plus show/hide link.
 *
 * @internal
 */
function routes_edit_table_header($class, $additional, $counter) {

  echo '<tr><td colspan="7"><a href="javascript:routes_group_showhide(\'' . $class . '\');" class="routes_group_showhide" id="' . $class . '">Show/hide rows ' . ($counter + 1) . ' to ...</a></td></tr>';

  echo '<tr class="' . $class . $additional . '">' .
    "<th>" . __("Order", 'routes') . "</th>" .
    "<th>" . __("City", 'routes') . "</th>" .
    "<th>" . __("State / Country", 'routes') . "</th>" .
    "<th>" . __("Latitude", 'routes') . "</th>" .
    "<th>" . __("Longitude", 'routes') . "</th>" .
    "<th>" . __("Delete?", 'routes') . "</th>" .
    "</tr>\n";
}

/**
 * Prints table header for POI section of post edit screen.
 *
 * @internal
 */
function routes_edit_poi_table_header($class, $additional, $counter) {

  echo '<tr><td colspan="4"><a href="javascript:routes_group_showhide(\'' . $class . '\');" class="routes_group_showhide" id="' . $class . '">Show/hide rows ' . ($counter + 1) . ' to ...</a></td></tr>';

  echo '<tr class="' . $class . $additional . '">' .
    "<th>" . __("Latitude", 'routes') . "</th>" .
    "<th>" . __("Longitude", 'routes') . "</th>" .
    "<th>" . __("Icon", 'routes') . "</th>" .
    "<th>" . __("Delete?", 'routes') . "</th>" .
    "</tr>\n";
}

/**
 * Prints an elevation calculation button on screen for post edit form.
 *
 * @internal
 */
function  routes_elev_calc_button($counter) {

  echo '<input type="button"' .
    ' name="routes_elev_btn' . $counter . '" id="routes_elev_btn' .
    $counter . '" value="' .
    __("Lookup", 'routes') . '" ' .
    ' onclick="routes_lookup_elev(' .
    "this.form.routes_lat$counter," .
    "this.form.routes_long$counter," .
    "this.form.routes_elev$counter);return true;" . '"' .
    "/>";
}

/**
 * Checks for existence of PHP 5.x XML library.
 *
 * @return bool
 *   TRUE if exists, FALSE if not.
 *
 * @internal
 */
function routes_php5_xml_lib_exists() {
  if (!function_exists('class_exists')) {
    return FALSE;
  }

  return class_exists('DOMDocument');
}

/**
 * Returns the numeric portion of the text.
 */
function routes_numpart($text) {
  $nonnums = "/[^\d.-]/";
  return preg_replace($nonnums, "", $text);
}

/**
 * Returns Y or N for 1/0.
 *
 * @internal
 *
 * @see routes_get_yn_option
 */
function routes_num_to_yn($num) {
  if ($num) {
    return "Y";
  }

  return "N";
}

/**
 * Escapes double quotes so that the text can be displayed as a formfield value.
 *
 * @internal
 */
function routes_fix_quotes_value($text) {
  return preg_replace('|"|', '&quot;', stripslashes($text));
}

/**
 * Prints out JavaScript for admin pages.
 */
function routes_js_admin_header() {
  global $routesLookupElev;
  global $editing;
  global $routesBlankLines;
  global $routesUseMetric;

  if (!$editing) {
     // we do not need any of this JS stuff if we are not on a post/page
     // edit screen
     return;
  }

  routes_print_leaflet();
  routes_print_icons();
?>
<script type="text/javascript">
//<![CDATA[

var routes_ext_info_shown = 0;
var routes_auto_elev = <?php echo $routesLookupElev; ?>;
var routes_use_metric = <?php echo $routesUseMetric; ?>;
var routes_blank_lines = <?php echo $routesBlankLines; ?>;
//]]>
</script>
<?php
    wp_enqueue_script('routes-admin-js');
}

/**
 * Prints out the JavaScript for footers of admin pages.
 */
function routes_js_admin_footer() {
  global $routesJSForFooter;
  global $editing;

  if (!$editing) {
     // we do not need any of this JS stuff if we are not on a post/page
     // edit screen
     return;
  }

?>
<script type="text/javascript">
//<![CDATA[
<?php
   echo $routesJSForFooter;
?>
//]]>
</script>
<?php
}

/**
 * Handles Ajax elevation requests.
 */
function routes_ajax_elevation_lookup() {
  global $routesUseMetric;

  $locations = new stdClass();
  $locations->locations = $_POST['points'];

  $ch = curl_init('https://api.open-elevation.com/api/v1/lookup');
  curl_setopt( $ch, CURLOPT_HTTPHEADER, array('Content-Type:application/json', 'Accept:application/json'));
  curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
  curl_setopt($ch, CURLOPT_FAILONERROR, TRUE);
  curl_setopt($ch, CURLOPT_POST, TRUE);
  curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode($locations));
  curl_setopt($ch, CURLOPT_TIMEOUT, 10);
  $result = @curl_exec($ch);
  curl_close($ch);
  if ($ok) {
    echo $result;
    wp_die();
  }

  // If we get here, the call to open-elevation failed. Try USGS instead,
  // even though this is only viable in the US (returns a negative elevation
  // if it's in an unsupported location). Also you have to call point by point.
  $usgs_output = new stdClass();
  $usgs_output->results = [];
  $usgs_data = [
    'output' => 'json',
  ];
  if ($routeUseMetric) {
    $usgs_data['units'] = 'Meters';
  }
  else {
    $usgs_data['units'] = 'Feet';
  }

  foreach ($locations->locations as $location) {
    $location = (array) $location;
    $usgs_data['y'] = $location['latitude'];
    $usgs_data['x'] = $location['longitude'];
    $ch = curl_init('http://ned.usgs.gov/epqs/pqs.php');
    curl_setopt( $ch, CURLOPT_HTTPHEADER, array('Content-Type:application/x-www-form-urlencoded', 'Accept:application/json'));
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
    curl_setopt($ch, CURLOPT_FAILONERROR, TRUE);
    curl_setopt($ch, CURLOPT_POST, TRUE);
    curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query($usgs_data));
    curl_setopt($ch, CURLOPT_TIMEOUT, 10);
    $output = @curl_exec($ch);
    curl_close($ch);
    if ($output && strpos($output, 'error') === FALSE) {
      $result = json_decode($output);
      $elevation = $result->USGS_Elevation_Point_Query_Service->Elevation_Query->Elevation;
      $item = new stdClass();
      $item->elevation = $elevation;
      $item->original = $output;
      $usgs_output->results[] = $item;
      if ($elevation < 0) {
        // Stop if we get a negative elevation, indicating error.
        break;
      }
    }
    else {
      // Stop if we get a Curl error.
      break;
    }
  }

  if (!count($usgs_output->results)) {
    // Both servers failed, so send an error output.
    $err_1 = new stdClass();
    $err_1->elevation = -2500;
    $err_1->last_query = http_build_query($usgs_data);
    $err_1->last_result = $output;
    $err_1->last_input = $location;
    $usgs_output->results = [$err_1];
  }
  echo json_encode($usgs_output);
  wp_die();
}
